#!/usr/bin/python3

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
import psycopg2, sys

myhost = 'localhost' 
#myhost = 'scischool.in'
mydb = 'dcqbank'
#myhost = '13.232.225.206'

MAXMOD = 8

class MainWindow(QMainWindow):   # Creating main window class
	dbname = None
	dbpasswd = None
	conn = None
	cursor = None
	subFilter = 'All'
	semFilter = 'All'
	courses = None
	courseCode = None
	courseName = None
	syllYear = None
	inst = None
	subject = None
	degree = None
	semester = None
	language = None
	qpModel = None
	duration = None
	numModules = None
	moduleMarkString = None

	syllYears = ['2020', '2021', '2022', '2023', '2024', '2025']
	institutions = ['AC', 'UD', 'DC']
	subjects = [ 'Mathematics', 'Computational Mathematics','EVS',\
	'Statistics', 'Physics','Chemistry','Botany','Zoology', 'Economics',\
	'Management Studies','Politics','History','Islamic History','Philosophy',\
	'Hindi','Sanskrit','Malayalam','Kannada','Arabic','Urdu','English',\
	'Functional English']
	subjectFilter = ['All'] + subjects
	degrees = ['BA', 'BSc', 'MA', 'MSc','FYIMP','FYUGP']
	semesters = ['First', 'Second', 'Third', 'Fourth', 'Fifth', 'Sixth', 'Seventh', 'Eighth']
	semesterFilter = ['All']+semesters
	languages = ['english', 'hindi', 'sanskrit','malayalam','kannada','urdu','arabic']
	qpModels = ['M50KU','M60KU','M70KU','M48KU','M80KU','M50CU','M70CU','M70DC']
	durations = ['90 min', '120 min', '180 min']
	qpMarks = {'M50CU':50,'M70CU':70,'M50KU':50,'M60KU':60,'M70KU':70,'M48KU':48,'M80KU':80}
	 
	def __init__(self, *args, **kwargs):
		super(MainWindow, self).__init__(*args, **kwargs)

		palette = QPalette()
		palette.setColor(QPalette.ColorRole(10), QColor(150,175,150)) 
		self.setPalette(palette)	
		self.setGeometry(100, 100, 1150, 700)			
		fixedfont = QFontDatabase.systemFont(QFontDatabase.FixedFont)
		fixedfont.setPointSize(10)

		layout = QVBoxLayout()	
		vbox1 = QVBoxLayout()
		layout.addLayout(vbox1)
		hbox1 = QHBoxLayout()
		vbox1.addLayout(hbox1)
		self.filterSubjectButton = QPushButton('Subject: All')
		hbox1.addWidget(self.filterSubjectButton)
		m = QMenu()
		for sub in self.subjectFilter:
			m.addAction(sub, lambda item=sub: self.setSubjectFilter(item))
		self.filterSubjectButton.setMenu(m)
		
		self.filterSemesterButton = QPushButton('Semester: All')
		hbox1.addWidget(self.filterSemesterButton)
		m = QMenu()
		for sub in self.semesterFilter:
			m.addAction(sub, lambda item=sub: self.setSemesterFilter(item))
		self.filterSemesterButton.setMenu(m)
	
		x = QLabel('Server')
		hbox1.addWidget(x)
		self.hostLineEdit = QLineEdit(myhost)
		self.hostLineEdit.setMaximumWidth(200)
		hbox1.addWidget(self.hostLineEdit)
		x = QLabel('Database')
		hbox1.addWidget(x)		
		self.dbLineEdit = QLineEdit(mydb)
		self.dbLineEdit.setMaximumWidth(200)
		hbox1.addWidget(self.dbLineEdit)

		self.dbConnectButton = QPushButton('Connect')
		s = 'Connects to the Database and Lists all the Courses found.\nTo restrict the number of items displayed\nuse the Subject and Semester filters'
		self.dbConnectButton.setToolTip(s)
		hbox1.addWidget(self.dbConnectButton)
		self.dbConnectButton.clicked.connect(self.connectDB)
		hbox1.addStretch()
		hb = QHBoxLayout()
		vbox1.addLayout(hb)
		self.qeditor = QPlainTextEdit()
		self.qeditor.setFont(fixedfont)
		self.qeditor.setLineWrapMode(1)
		hb.addWidget(self.qeditor)
		hb = QHBoxLayout()
		vbox1.addLayout(hb)
		x = QLabel('To Create a Course Provide all details. To Delete Couse Code is enough')
		x.setStyleSheet("background-color:white")
		x.setAlignment(Qt.AlignCenter)
		hb.addWidget(x)

		hbox4 = QHBoxLayout()
		vbox1.addLayout(hbox4)	

		self.syllYearButton = QPushButton('Syllabus Year')
		hbox4.addWidget(self.syllYearButton)
		m = QMenu()
		for s in self.syllYears:
			m.addAction(s, lambda item=s: self.selectYear(item))
		self.syllYearButton.setMenu(m)

		self.instButton = QPushButton('Institution')
		hbox4.addWidget(self.instButton)
		m = QMenu()
		for s in self.institutions:
			m.addAction(s, lambda item=s: self.selectInst(item))
		self.instButton.setMenu(m)

		self.subjectButton = QPushButton('Subject')
		hbox4.addWidget(self.subjectButton)
		m = QMenu()
		for sub in self.subjects:
			m.addAction(sub, lambda item=sub: self.selectSubject(item))
		self.subjectButton.setMenu(m)
		
		self.degreeButton = QPushButton('Degree')
		hbox4.addWidget(self.degreeButton)
		m = QMenu()
		for deg in self.degrees:
			m.addAction(deg, lambda item=deg: self.selectDegree(item))
		self.degreeButton.setMenu(m)


		self.semesterButton = QPushButton('Semester')
		hbox4.addWidget(self.semesterButton)
		m = QMenu()
		for s in self.semesters:
			m.addAction(s, lambda item=s: self.selectSemester(item))
		self.semesterButton.setMenu(m)

		self.languageButton = QPushButton('Language')
		hbox4.addWidget(self.languageButton)
		m = QMenu()
		for s in self.languages:
			m.addAction(s, lambda item=s: self.selectLanguage(item))
		self.languageButton.setMenu(m)
		
		self.qpModelButton = QPushButton('QP Model')
		hbox4.addWidget(self.qpModelButton)
		m = QMenu()
		for s in self.qpModels:
			m.addAction(s, lambda item=s: self.selectQpModel(item))
		self.qpModelButton.setMenu(m)
		
		self.durationButton = QPushButton('Duration')
		hbox4.addWidget(self.durationButton)
		m = QMenu()
		for s in self.durations:
			m.addAction(s, lambda item=s: self.selectDuration(item))
		self.durationButton.setMenu(m)

		hbox5 = QHBoxLayout()
		vbox1.addLayout(hbox5)
		spacer = QLabel(' '*34)
		hbox5.addWidget(spacer)
		self.numModulesButton = QPushButton('numModules')
		self.numModulesButton.setVisible(False)
		hbox5.addWidget(self.numModulesButton)
		m = QMenu()
		for s in  ['4', '5', '6','7', '8']:m.addAction(s, lambda item=s: self.selectnumModules(item))
		self.numModulesButton.setMenu(m)

		self.pmLabels = [None]*MAXMOD
		self.pmEdits = [None]*MAXMOD
		for k in range(MAXMOD):
			self.pmLabels[k] = QLabel('M%d Mark->'%(k+1))
			self.pmLabels[k].setVisible(False)
			hbox5.addWidget(self.pmLabels[k])
			self.pmEdits[k] = QLineEdit('0')
			self.pmEdits[k].setMaximumWidth(50)
			self.pmEdits[k].setVisible(False)
			hbox5.addWidget(self.pmEdits[k])
		hbox5.addStretch()

		x = QLabel('To Add a Course provide all the information. To remove,  courseCode is enough.')	
		hbox3 = QHBoxLayout()
		layout.addLayout(hbox3)

		f = QFormLayout()
		vbox1.addLayout(f)
		self.codeLineEdit = QLineEdit('')
		f.addRow(QLabel("Course Code"), self.codeLineEdit)

		self.pass1LineEdit = QLineEdit('')
		self.pass1LineEdit.setEchoMode(QLineEdit.Password)
		f.addRow(QLabel("Course password"), self.pass1LineEdit)

		self.pass2LineEdit = QLineEdit('')
		self.pass2LineEdit.setEchoMode(QLineEdit.Password)
		f.addRow(QLabel("Course password"), self.pass2LineEdit)

		self.nameLineEdit = QLineEdit('')
		f.addRow(QLabel("Course Name"), self.nameLineEdit)

		hbox5 = QHBoxLayout()
		vbox1.addLayout(hbox5)
		self.courseButton = QPushButton('Select Existing Course')
		s = 'Select a course from the pull down menu\nYou may delete or modify the selected course.\nYou can also create a new course after editing the courseCode'
		self.courseButton.setToolTip(s)
		self.courseButton.clicked.connect(self.selectCourse)
		hbox5.addWidget(self.courseButton)
		self.courseMenu = QMenu()
		self.courseButton.setMenu(self.courseMenu)
		self.modButton = QPushButton('Modify Selected Course')
		s = 'Modify the selected course.\nCourse code,QP model and numModules cannot be changed.\nLeave the Password field empty if you do not want to modify it'
		self.modButton.setToolTip(s)
		self.modButton.clicked.connect(self.modifyCourse)
		hbox5.addWidget(self.modButton)
		self.courseButton.setMenu(self.courseMenu)
		self.delButton = QPushButton('Delete Selected Course')
		s = 'Delete the selected Course'
		self.delButton.setToolTip(s)
		self.delButton.clicked.connect(self.delCourse)
		hbox5.addWidget(self.delButton)
		self.addButton = QPushButton('Create New Course')
		s = 'Create a New Course using the Information\n from the text Fields and Menus above'
		self.addButton.setToolTip(s)
		self.addButton.clicked.connect(self.addCourse)
		hbox5.addWidget(self.addButton)
		hbox5.addStretch()
		
		self.webPassButton = QPushButton('Set Web password')
		s = 'Change the password for the Web based Question paper generation'
		self.webPassButton.setToolTip(s)
		self.webPassButton.clicked.connect(self.setWebPassword)
		hbox5.addWidget(self.webPassButton)
		self.webPass1LineEdit = QLineEdit('')
		self.webPass1LineEdit.setEchoMode(QLineEdit.Password)
		hbox5.addWidget(self.webPass1LineEdit)
		self.webPass2LineEdit = QLineEdit('')
		self.webPass2LineEdit.setEchoMode(QLineEdit.Password)
		hbox5.addWidget(self.webPass2LineEdit)

		hb = QHBoxLayout()
		layout.addLayout(hb)	
		self.msg = QLabel('')
		hb.addWidget(self.msg)
		
		container = QWidget()
		container.setLayout(layout)
		self.setCentralWidget(container)  	# making it central widget	
		self.show()		
		# End of __init__

	def getPassword(self):
		text, ok = QInputDialog.getText(None, "Authenticate", "Password for Database", QLineEdit.Password)
		if ok and text:
			return text			

	def dialog_critical(self, s):
		dlg = QMessageBox(self)
		dlg.setText(s)
		dlg.setIcon(QMessageBox.Critical)
		dlg.show()			
	
	def connectDB(self):
		if self.conn != None:
			self.conn.close()
			self.conn = None
		self.cursor = None
		host = self.hostLineEdit.text()
		db = self.dbLineEdit.text()
		pw = self.getPassword()
		self.qeditor.setPlainText('')	
		try:
			print('Connecting to %s on %s ...'%(db,host), end =' ')	
			sys.stdout.flush()	
			self.conn = psycopg2.connect(database=db, host=host, user= db, password=pw)
			self.cursor = self.conn.cursor()
			self.dbname = db	
			print ('Fetched data\n')
			sys.stdout.flush()					
		except:
			s = 'Connection to Database %s failed.'%db
			self.dialog_critical(s)		
			return		
		self.updateCourseList()		

	def updateCourseList(self):
		try:
			sql =  'select * from coursedetails order by code;'
			self.cursor.execute(sql)
			result = self.cursor.fetchall()
		except:
			s = 'Fetching Course details from Database %s failed.'%self.dbname
			self.dialog_critical(s)		
			return				
		self.courses = result		# Store the course list information
		self.displayCourses()
		self.msg.setText('Found %d Courses.'%len(result))

	def setSubjectFilter(self,s):
		self.subFilter = s
		self.filterSubjectButton.setText('Subject:'+s)
		self.displayCourses()

	def setSemesterFilter(self,s):
		self.semFilter = s
		self.filterSemesterButton.setText('Semester:'+s)
		self.displayCourses()

	def displayCourses(self):
		txt = ''
		self.qeditor.setPlainText(txt)		
		txt = '%-20s %-6s %-4s %-30s %-10s %-6s %-8s %-10s %-10s %-10s %-10s\n'%('Course Code','Year','Inst','Course Name','Subject','Degree','Semester','TexLang','QPModel', 'Duration', 'Modules')

		self.courseMenu.clear()
		if self.courses == None:
			return
		for row in self.courses:
			if self.subFilter != 'All' and self.subFilter != row[4]:  	
				continue
			if self.semFilter != 'All' and self.semFilter != row[6]:  	
				continue
			self.courseMenu.addAction(row[0][8:], lambda item=row[0]: self.selectCourse(item))
			
			s = '%-20s %-6s %-4s %-30s %-10s %-6s %-8s %-10s %-10s %-10s %-10s\n'%(row[0][:],row[1],row[2],row[3],row[4],row[5],row[6],row[7],row[8],row[9],row[10])
			txt += s[:-1] + '\n'  
		self.qeditor.setPlainText(txt)
		self.qeditor.setReadOnly(True)
						
	def setWebPassword(self):
		if self.cursor == None:
			self.dialog_critical('Not Connected to any Database.')
			return			
		wpw = self.webPass1LineEdit.text()
		if len(wpw) < 6:
			self.dialog_critical('Must have minimum 6 characters.')
			return	
		if wpw != self.webPass2LineEdit.text():
			self.dialog_critical('Passwords not matching.')
			return
		sql = "alter role reader with password '%s';"%wpw
		try:
			self.cursor.execute(sql)	
			self.conn.commit()		
			self.dialog_critical('Web Password changed.')
		except:
			self.dialog_critical('Web Password changing failed.')
		
	def selectYear(self,s):
		self.syllYear = s
		self.syllYearButton.setText(s)

	def selectInst(self,s):
		self.inst = s
		self.instButton.setText(s)

	def selectSubject(self,s):
		self.subject = s
		self.subjectButton.setText(s)
				
	def selectDegree(self,s):
		self.degree = s
		self.degreeButton.setText(s)			

	def selectSemester(self,s):
		self.semester = s
		self.semesterButton.setText(s)

	def selectLanguage(self,s):
		self.language = s
		self.languageButton.setText(s)

	def selectQpModel(self,s):
		self.qpModel = s
		self.qpModelButton.setText(s)
		self.numModulesButton.setVisible(True)
		self.numModulesButton.setText('numModules')
		self.numModules = None		# Reset number of modules
		for k in range(MAXMOD):
			self.pmLabels[k].setVisible(False)
			self.pmEdits[k].setVisible(False)

	def selectDuration(self,s):
		self.duration = s
		self.durationButton.setText(s)

	def selectnumModules(self,s):		# Also controls the mark widgets
		self.numModulesButton.setText(s+' Modules')
		self.numModules = eval(s)
		if self.qpModel not in ['M50CU', 'M70CU']:
			return
		for k in range(MAXMOD):			# Show the mark entry widgets
			if(k < self.numModules):
				self.pmLabels[k].setVisible(True)
				self.pmEdits[k].setVisible(True)
			else:
				self.pmLabels[k].setVisible(False)
				self.pmEdits[k].setVisible(False)
		#self.getModuleMarks()

	def getModuleMarks(self):			# make the string from the Widget inputs
		if self.qpModel == None:
			self.dialog_critical('Error: QP Model is NOT selected')
			return 
		if self.qpModel not in ['M50CU', 'M70CU']:	# Unequal marks set by user
			self.moduleMarkString = '%d'%self.numModules
			for k in range(self.numModules):
				self.moduleMarkString += ':0'
			return True
		ss = '%d'%self.numModules
		tm = 0
		try:
			for k in range(self.numModules):
				ss += ':%s'%self.pmEdits[k].text()
				tm += eval(self.pmEdits[k].text())
		except:
			self.dialog_critical('Invalid marks given.')
			return
		n = self.qpMarks[self.qpModel]	
		if tm != n:
			self.dialog_critical('Section Marks total should be %d'%n)
			return	
		self.moduleMarkString = ss
		return True

	def initModuleMarks(self,ss):
			if self.qpModel not in ['M50', 'M70','M50CU', 'M70CU']:
				return
			sss = ss.split(':')	
			self.numModules = eval(sss[0])
			marks = sss[1:]
			if len(marks) != self.numModules:
				self.dialog_critical('numModules & marks NOT proper. %s'%ss)
				return
			try:
				for k in range(self.numModules):
					self.pmEdits[k].setText(marks[k])
			except:
				self.dialog_critical('Section marks NOT proper.')
				return
			return
			
	#----------------------- Select, Add & Delete ---------------------
	def selectCourse(self, code):
		for c in self.courses:
			if code == c[0]:
				self.courseCode = code
				self.pass1LineEdit.setText('')
				self.pass2LineEdit.setText('')
				self.codeLineEdit.setText(c[0][8:])
				self.syllYear = c[1]
				self.syllYearButton.setText(self.syllYear)
				self.inst = c[2]
				self.instButton.setText(self.inst)	
				self.nameLineEdit.setText(c[3])	
				self.subject = c[4]
				self.subjectButton.setText(self.subject)	
				self.degree = c[5]
				self.degreeButton.setText(self.degree)
				self.semester = c[6]
				self.semesterButton.setText(self.semester)
				self.language = c[7]
				self.languageButton.setText(self.language)
				self.qpModel = c[8]
				self.qpModelButton.setText(self.qpModel)
				self.selectQpModel(self.qpModel)
				self.duration = c[9]
				self.durationButton.setText(self.duration)	
				self.moduleMarkString = c[10]
				np = self.moduleMarkString.split(':')[0]			
				self.numModules = eval(np)
				self.selectnumModules(np)		# control the widgets
				self.numModulesButton.setText('%d Modules'%self.numModules)		
				self.initModuleMarks(self.moduleMarkString)		# Sets the mark LineEdit values
				
	def validate(self):
		if self.cursor == None:
			self.dialog_critical('Not Connected to any Database')
			return False	
		if self.syllYear == None:		# Year & Inst prefixed with courseCode
			self.dialog_critical('Syllabus Year not selected.')
			return False
		if self.inst == None:		# Year & Inst prefixed with courseCode
			self.dialog_critical('Institution type not selected.')
			return False
		code = self.codeLineEdit.text().strip()
		if len(code) < 6:
			self.dialog_critical('Course Code too short or empty.')
			return False
		spchars = '!#@$%^&*()_\/?'
		for c in spchars:
			if c in code:
				self.dialog_critical('Letter %s not allowed in Coursecode.'%c)
				return False
		name = self.nameLineEdit.text()
		if len(name.strip()) < 5:
			self.dialog_critical('Course Name too short or not given.')
			self.courseName = None
			return False	
		for c in spchars:
			if c in name:
				self.dialog_critical('Letter %s not allowed in Course name.'%c)
				return False

		self.courseName = name			
		if self.subject == None:
			self.dialog_critical('Subject not selected.')
			return False	
		if self.degree == None:
			self.dialog_critical('Degree not selected.')
			return False	
		if self.semester == None:
			self.dialog_critical('Semester not selected.')
			return False	
		if self.language == None:
			self.dialog_critical('Type setting language not selected.')
			return False	
		if self.qpModel == None:
			self.dialog_critical('QP Model not selected.')
			return False	
		if self.duration == None:
			self.dialog_critical('Duration not selected.')
			return False	
		if self.numModules == None:
			self.dialog_critical('Number of Modules not selected.')
			return False	
		if self.getModuleMarks() != True:
			self.dialog_critical('Module Marks not obtained.')
			return False	
		
	def modifyCourse(self):	
		if self.validate() == False:
			return
		code = 'Y' + self.syllYear + self.inst + '_' +self.codeLineEdit.text().strip()
		if code != self.courseCode:
			self.dialog_critical('You cannot modify the Course Code.')
			return False	
		try:		
			sql = "update coursedetails set name = '%s', subject = '%s', degree = '%s', syllabusyear = '%s', semester = '%s', texlanguage = '%s', duration = '%s', inst= '%s' where code = '%s';"%(self.courseName, self.subject, self.degree, self.syllYear, self.semester, self.language, self.duration, self.inst, self.courseCode)
			self.cursor.execute(sql)
			self.conn.commit()
			self.updateCourseList()
		except:
			self.dialog_critical('Failed to Modify Course %s.'%code)

		cpw1 = self.pass1LineEdit.text()
		cpw2 = self.pass2LineEdit.text()
		if cpw1 == '' and cpw2 == '':
			return
		if len(cpw1) < 6:
			self.dialog_critical('Course password must have minimum 6 characters.')
			return False	
		if cpw1 != cpw2:
			self.dialog_critical('Course passwords not matching.')
			return False	
		
		sql = "alter user %s with password '%s'"%(code.lower(), cpw1)
		try:
			self.cursor.execute(sql)
			self.conn.commit()
		except:
			self.dialog_critical('Failed to Modify password.')
			
		self.msg.setText('Course Password Changed')
				
	def addCourse(self):	
		if self.validate() == False:	# check fields except the password
			return
		cpw = self.pass1LineEdit.text()
		if len(cpw) < 6:
			self.dialog_critical('Course password must have minimum 6 characters.')
			return False	
		if cpw != self.pass2LineEdit.text():
			self.dialog_critical('Course passwords not matching.')
			return False	
			
		code = 'Y' + self.syllYear + self.inst + '_' +self.codeLineEdit.text().strip()	
		try:
			sql = "CREATE TABLE %s(qtext VARCHAR, atext VARCHAR, author VARCHAR, modnum int, qtype CHAR(1), qnum int, relations VARCHAR, mark int, status VARCHAR,figname VARCHAR, figure bytea);"%code
			#print (sql)
			self.cursor.execute(sql)
			sql = "create user %s with password '%s'"%(code, cpw)
			self.cursor.execute(sql)
			sql = 'grant all on table %s to %s'%(code,code)
			self.cursor.execute(sql)
			self.sql = 'grant select on table %s to reader'%(code)
			self.cursor.execute(sql)
			sql = "insert into coursedetails (code, name, subject, degree, syllabusyear, semester, texlanguage, qpmodel, duration,nmodules, inst) values ('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s','%s','%s');"%(code, self.courseName, self.subject, self.degree, self.syllYear, self.semester, self.language, self.qpModel, self.duration, self.moduleMarkString,self.inst)
			self.cursor.execute(sql)
			self.conn.commit()
			self.updateCourseList()
		except:
			self.dialog_critical('Failed to add Course %s.'%code)
				
	def delCourse(self):
		code = 'Y' + self.syllYear + self.inst + '_' +self.codeLineEdit.text().strip()
		reply = QMessageBox.question(GUI, 'Delete Course', 'Proceed with Delete ?',QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
		if reply == QMessageBox.No:
			return	
		try:
			sql = "delete from coursedetails where code = '%s'"%code
			self.cursor.execute(sql)		
			sql = "drop table %s;"%code
			self.cursor.execute(sql)		
			sql = "drop user %s;"%code
			self.cursor.execute(sql)
			self.conn.commit()
		except:
			self.dialog_critical('Error: Delete Course %s'%code)
		self.updateCourseList()
				
app = QApplication(sys.argv)   # Main program
GUI = MainWindow()
app.exec_()		# main loop

